"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
var path_1 = require("path");
var express_1 = __importDefault(require("express"));
var webpack_1 = __importStar(require("webpack"));
var webpack_dev_middleware_1 = __importDefault(require("webpack-dev-middleware"));
var webpack_hot_middleware_1 = __importDefault(require("webpack-hot-middleware"));
var webpack_node_externals_1 = __importDefault(require("webpack-node-externals"));
var flags = {
    'server-config': {
        optionKey: 'serverConfigPath',
        code: 's',
        helpMsg: 'Specify path to your webpack config for the server',
        needsValue: true,
    },
    'client-config': {
        optionKey: 'clientConfigPath',
        code: 'c',
        helpMsg: 'Specify path to your webpack config for client assets',
        needsValue: true,
    },
    'help': {
        optionKey: 'help',
        code: 'h',
        helpMsg: 'Display help',
    },
};
var requireRef = require;
var exit = function (exitCode) {
    if (process.env.NODE_ENV !== 'test') {
        process.exit(exitCode);
    }
    console.warn('exited', exitCode);
};
var flagCodeMap = Object.keys(flags).reduce(function (map, flagKey) {
    var _a;
    return (__assign(__assign({}, map), (_a = {}, _a[flags[flagKey].code] = flags[flagKey], _a)));
}, {});
var parseParamToken = function (token) {
    var match = /(-([\w-]{1,1})|--([\w-]+))(?:=(.*))?$/.exec(token);
    if (!match)
        return [{}];
    var _a = match.filter(Boolean), flag = _a[2], value = _a[3];
    var flagSpec = flags[flag && flag.toLowerCase()] || flagCodeMap[flag];
    if (!flagSpec)
        throw new Error("Invalid parameter " + flag);
    return [flagSpec, value];
};
var fatalError = function (msg) {
    process.stdout.write(msg + "\n");
    exit(1);
};
exports.parseCLIParams = function (args) {
    var skip;
    var flagSet;
    return args.reduce(function (options, arg, index) {
        var _a, _b;
        if (skip) {
            skip = false;
            return options;
        }
        var getNextValue = (function () {
            skip = true;
            return args[index + 1];
        });
        var _c = parseParamToken(arg), _d = _c[0], flag = _d.optionKey, needsValue = _d.needsValue, value = _c[1];
        if (flag) {
            flagSet = true;
            if (value)
                return __assign(__assign({}, options), (_a = {}, _a[flag] = value, _a));
            return __assign(__assign({}, options), (_b = {}, _b[flag] = (needsValue ? getNextValue() : true), _b));
        }
        if (!flagSet && index === args.length - 1) {
            options.entryConfigPath = arg;
        }
        return options;
    }, {});
};
var state = {
    startCount: 0,
    indexPos: 0,
    hookedExpressApp: undefined,
};
var routers = [];
var propertyHook = function (expressApp, keyName) {
    var indexPos = state.indexPos;
    switch (keyName) {
        case 'listen':
            if (state.startCount > 1)
                return function () { };
            return expressApp[keyName];
        case 'use': {
            state.indexPos += 1;
            routers[indexPos] = express_1.default.Router();
            if (state.startCount === 1) {
                expressApp.use(function () {
                    var args = [];
                    for (var _i = 0; _i < arguments.length; _i++) {
                        args[_i] = arguments[_i];
                    }
                    return routers[indexPos].apply(routers, args);
                });
            }
            return function () {
                var _a;
                var args = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    args[_i] = arguments[_i];
                }
                return (_a = routers[indexPos]).use.apply(_a, args);
            };
        }
        default:
            return expressApp[keyName];
    }
};
var getConstructorHook = function (express, clientConfig) {
    return function (_target, thisArg, argList) {
        state.startCount += 1;
        state.indexPos = 0;
        if (state.hookedExpressApp)
            return state.hookedExpressApp;
        var expressApp = express.apply(void 0, argList);
        if (clientConfig) {
            var clientCompiler = webpack_1.default(clientConfig);
            var devServerOptions = clientConfig.devServer;
            expressApp.use(webpack_hot_middleware_1.default(clientCompiler));
            expressApp.use(webpack_dev_middleware_1.default(clientCompiler, devServerOptions));
        }
        state.hookedExpressApp = new Proxy(expressApp, { get: propertyHook });
        return state.hookedExpressApp;
    };
};
var printHelp = function () {
    var maxLen = Object.keys(flags).reduce(function (max, key) { return (key.length > max ? key.length : max); }, 0);
    var flagHelp = Object.keys(flags).reduce(function (flagHelp, flagKey) {
        var space = ' '.repeat(Math.max(maxLen - flagKey.length - 1, 0));
        var _a = flags[flagKey], code = _a.code, helpMsg = _a.helpMsg;
        var codeMsg = code ? "-" + code + ", " : '';
        var helpLine = codeMsg + "--" + flagKey + "\t\t" + space + helpMsg;
        return __spreadArrays(flagHelp, [helpLine]);
    }, []);
    var msgData = __spreadArrays([
        'Usage: webpack-express-dev-middleware [script]',
        '\twebpack-express-dev-middleware [options]',
        '\twedm [options]',
        '',
        'Options:'
    ], flagHelp, [
        '\n',
    ]);
    var msg = msgData.reduce(function (msg, line) { return msg + "\n" + line; });
    process.stdout.write(msg);
    exit();
};
var identifyEntries = function (entryPath) {
    //TODO: unit test this
    //TODO: is this even needed? the user is typing this in command line they can just write the correct thing?
    var isStablePath = function (path) {
        if (!path)
            return false;
        if (path[0] === '~')
            return true;
        return path_1.isAbsolute(path);
    };
    var reduceEntryArray = function (entryConfig, path) {
        if (!Array.isArray(entryConfig))
            return undefined;
        if (entryConfig.length > 2) {
            throw new Error("Too many webpack configurations contained in " + path);
        }
        var entries = entryConfig.reduce(function (entries, entry) {
            if (entry.target === 'node')
                return __assign(__assign({}, entries), { serverConfig: entry });
            return __assign(__assign({}, entries), { clientConfig: entry });
        }, {});
        if (!entries.serverConfig) {
            throw new Error("No server config was found in " + path);
        }
        return entries;
    };
    var fullPath = isStablePath(entryPath) ? entryPath : path_1.join(process.cwd(), entryPath);
    var entryConfig = fullPath && require(fullPath);
    var webpackConfigs = reduceEntryArray(entryConfig, fullPath);
    if (webpackConfigs)
        return webpackConfigs;
};
//TODO unit test this stuff applies when its supposed to
var applyPreferredConfigSettings = function (_a) {
    var serverConfig = _a.serverConfig, clientConfig = _a.clientConfig;
    if (serverConfig) {
        serverConfig.mode = 'development';
        serverConfig.watch = true;
        if (!serverConfig.externals) {
            serverConfig.externals = [webpack_node_externals_1.default()];
        }
        else if (Array.isArray(serverConfig.externals)) {
            var nodeExt = serverConfig.externals.find(function (item) {
                if (typeof item !== 'function')
                    return;
                return /mark this module as external/.exec(item.toString());
            });
            if (!nodeExt) {
                serverConfig.externals.push(webpack_node_externals_1.default());
            }
        }
    }
    if (clientConfig) {
        var plugins = clientConfig.plugins, entry = clientConfig.entry;
        var hmrEntry = 'webpack-hot-middleware/client';
        clientConfig.mode = 'development';
        if (!plugins) {
            plugins = [new webpack_1.HotModuleReplacementPlugin()];
            clientConfig.plugins = plugins;
        }
        if (!plugins.find(function (item) { return item instanceof webpack_1.HotModuleReplacementPlugin; })) {
            plugins.unshift(new webpack_1.HotModuleReplacementPlugin());
        }
        if (typeof entry === 'string') {
            entry = [hmrEntry, entry];
            clientConfig.entry = entry;
        }
        if (Array.isArray(entry) && !entry.find(function (item) { return item.match(/webpack-hot-middleware/); })) {
            entry.unshift(hmrEntry);
        }
    }
};
exports.resolveConfigs = function (cmdLineArgs) {
    var _a = exports.parseCLIParams(cmdLineArgs), serverConfigPath = _a.serverConfigPath, clientConfigPath = _a.clientConfigPath, entryConfigPath = _a.entryConfigPath, help = _a.help;
    var configs = {};
    if (help) {
        printHelp();
    }
    configs.serverConfig = serverConfigPath && require(serverConfigPath);
    configs.clientConfig = clientConfigPath && require(clientConfigPath);
    if (entryConfigPath) {
        var _b = identifyEntries(entryConfigPath), serverConfig = _b.serverConfig, clientConfig = _b.clientConfig;
        configs.serverConfig = serverConfig;
        configs.clientConfig = clientConfig;
    }
    if (!configs.serverConfig) {
        printHelp();
        fatalError('No webpack config was specified for the server');
    }
    if (!configs.clientConfig) {
        console.warn('No client config was specified');
    }
    applyPreferredConfigSettings(configs);
    return configs;
};
var getHookedRequire = function (clientConfig) {
    return function (modName) {
        var mod = requireRef(modName);
        if (modName !== 'express')
            return mod;
        return new Proxy(mod, { apply: getConstructorHook(mod, clientConfig) });
    };
};
exports.runServer = function (args) {
    var _a = exports.resolveConfigs(args), serverConfig = _a.serverConfig, clientConfig = _a.clientConfig;
    var hookedRequire = getHookedRequire(clientConfig);
    return new Promise(function (resolve) {
        webpack_1.default(serverConfig, function (err, _a) {
            var compilation = _a.compilation;
            try {
                var assets_1 = compilation.assets, errors = compilation.errors;
                if (err) {
                    console.error(err);
                    return;
                }
                if (errors.length > 0) {
                    errors.forEach(function (err) {
                        console.error(err);
                    });
                    throw new Error('Could not compile server');
                }
                var serverCode = Object.keys(assets_1).reduce(function (src, assetKey) { return ("" + src + assets_1[assetKey].source()); }, '');
                var runServerCode = new Function('require', serverCode);
                runServerCode(hookedRequire);
                resolve();
            }
            catch (err) {
                console.error(err);
            }
        });
    });
};
//# sourceMappingURL=lib.js.map